#include "strategy.h"

using namespace std;

//! get distance^2 from center of House
float get_dist(float x, float y)
{
	return pow(x - TEE_X, 2) + pow(y - TEE_Y, 2);
}

//! get distance^2 of two cordinates
float get_dist(float x1, float y1, float x2, float y2)
{
	return pow(x1 - x2, 2) + pow(y1 - y2, 2);
}

//! is a Stone in House
bool is_in_House(float x, float y)
{
	if (get_dist(x, y) < pow(HOUSE_R + STONE_R, 2)) {
		return true;
	}
	else {
		return false;
	}
}

//! sort Shot number (rank[] = {0, 1, 2 ... 15})
//  by distance from center of House (TEEX, TEEY)
void get_ranking(int *rank, const GAMESTATE* const gs)
{
	// init array
	for (int i = 0; i < 16; i++) {
		rank[i] = i;
	}

	// sort
	int tmp;
	for (int i = 1; i < gs->ShotNum; i++) {
		for (int j = i; j > 0; j--) {
			if (get_dist(gs->body[rank[j]][0], gs->body[rank[j]][1]) < get_dist(gs->body[rank[j - 1]][0], gs->body[rank[j - 1]][1])) {
				// swap
				tmp = rank[j];
				rank[j] = rank[j - 1];
				rank[j - 1] = tmp;
			}
			else {
				break;
			}
		}
	}
}

//! get best Shot to takeout the objective stone
void CreateTakeoutShot(const GAMESTATE* const gs, unsigned int num_target, SHOTVEC *vec_ret)
{
	/*** this is a sample of 'Simulation' function ***/

	const unsigned int shot_variations = 2;

	SHOTPOS target;
	SHOTVEC vec[shot_variations];

	// get position of target stone
	target.x = gs->body[num_target][0];
	target.y = gs->body[num_target][1];

	// get a HitShot (right Curl)
	target.angle = 0;
	CreateHitShot(target, 16, &(vec[0]));

	// get a HitShot (left Curl)
	target.angle = 1;
	CreateHitShot(target, 16, &(vec[1]));

	// try a shot 100 times
	const unsigned int trials = 100;
	unsigned int count[shot_variations] = { 0, 0 };

	// create temporary GAMESTATE
	GAMESTATE *gstmp;
	gstmp = (GAMESTATE *)malloc(sizeof(GAMESTATE));

	// simulate each shot
	for (unsigned int i = 0; i < trials; i++) {
		for (int j = 0; j < shot_variations; j++) {
			// copy current state to temporary state
			memcpy_s(gstmp, sizeof(GAMESTATE), gs, sizeof(GAMESTATE));

			// Simulation
			Simulation(gstmp, vec[j], RANDOM, NULL, -1);

			// check objective stone was in House or not
			// NOTE: 'Simulation' rewites gstmp as a state after the simulation
			if ( !is_in_House(gstmp->body[num_target][0], gstmp->body[num_target][1])) {
				count[j]++;
			}
		}
	}

	// calculate accuracy of Takeout Shot
	unsigned int best_num = 0;
	unsigned int count_max = trials + 1;
	for (int i = 0; i < shot_variations; i++) {
		if (count_max < count[i]) {
			count_max = count[i];
			best_num = i;
		}


		// NOTE: use cerr if you want output something on command line
		cerr << "shot[" << i << "] (" << count[i] << "/" << trials << ")" << endl;
	}

	cerr << "best_num = " << best_num << endl;

	memcpy_s(vec_ret, sizeof(SHOTVEC), &(vec[best_num]), sizeof(SHOTVEC));
}

//! choose best Shot
//  This sample choose:
//    1.Takeout Shot: hit and push the objective Stone away from House 
//      - IF No.1 Stone (the nearest Stone from center of House) is opponent Stone.
//    2.Guard Shot: put a Stone as hiding the objective Stone.
//      - IF No.1 Stone is my Stone.
//    3.Tee Shot: put a Stone on center of House
//      - IF there is no Stone in House
void getBestShot(const GAMESTATE* const gs, SHOTVEC *vec_ret)
{
	SHOTPOS pos;

	// ranking of Shot number
	// rank[n] = x;
	//   n : the n th Stone from the center of House
	//   x : the x th Shot in this End (corresponding to the number of GAMESTATE->body[x])
	int rank[16];

	// sort by distane from center
	get_ranking(rank, gs);

	// create Shot according to condition of No.1 Stone
	if (is_in_House(gs->body[rank[0]][0], gs->body[rank[0]][1]))
	{
		// get position of the objective Stone
		pos.x = gs->body[rank[0]][0];
		pos.y = gs->body[rank[0]][1];

		if (rank[0] % 2 != gs->ShotNum % 2) {
			// choose Shot 1.
			if (0) {
				// without simulation
				if (gs->body[rank[0]][0] < TEE_X) {
					pos.angle = 1;
				}
				else {
					pos.angle = 0;
				}

				// generate a Shot(SHOTVEC) by CreateHitShot (pass the objective position)
				CreateHitShot(pos, 16, vec_ret);
			}
			else {
				// with simulation
				CreateTakeoutShot(gs, rank[0], vec_ret);
			}
		}
		else {
			// choose Shot 2.
			// generate a Shot by CreateShot (put a Stone on objective position: pos)
			pos.y += 2.0;
			if (gs->body[rank[0]][0] < TEE_X) {
				pos.angle = 0;
			}
			else {
				pos.angle = 1;
			}

			CreateShot(pos, vec_ret);
		}
	}
	else {
		// choose Shot 3.

		// set the objective position as center of House
		pos.x = TEE_X;
		pos.y = TEE_Y;
		pos.angle = 0;

		// generate a Shot by CreateShot (put a Stone on objective position: pos)
		CreateShot(pos, vec_ret);
	}
}